from flask import Flask, render_template
from flask_orjson import OrjsonProvider

from app.extensions import db


def load_blueprints(flask_app):
    from app.rpc import rpc

    flask_app.register_blueprint(rpc)


def load_models(flask_app, flask_db):
    from app.models import clients

    flask_db.init_app(flask_app)

    with flask_app.app_context():
        flask_db.create_all()


def create_app():
    app = Flask(__name__, static_folder=None, template_folder="templates")
    app.config["SQLALCHEMY_DATABASE_URI"] = "sqlite:///db.sqlite"
    app.secret_key = "secret"
    app.json = OrjsonProvider(app)

    load_models(app, db)
    load_blueprints(app)

    @app.route("/")
    def index():
        return render_template("index.html")

    return app



from flask import Blueprint


def load_nested_blueprints(bp):
    from .auth import auth
    from .clients import clients
    from .tester import bp as tester

    bp.register_blueprint(auth)
    bp.register_blueprint(clients)
    bp.register_blueprint(tester)


rpc = Blueprint("rpc", __name__, url_prefix="/rpc")

load_nested_blueprints(rpc)



from flask_rpc.latest import RPCResponse as Res


# accept list return int
def add_numbers(data: list):
    return Res.success(data=sum(data), message="add_numbers")


# accept string return string
def add_string(data: str):
    return Res.success(data=f"Hello {data}!", message="add_string")



import typing as t


class RPCRequest:
    @classmethod
    def build(
        cls,
        function: str,
        data: t.Union[
            str, int, float, bool, t.List[t.Any], t.Dict[str, t.Any], None
        ] = None,
    ) -> t.Dict[str, t.Any]:
        """
        Build a request.

        Version 1.0.

        :param function: Str
        :param data: Any (JSON serializable)
        :return:
        """
        return {"wrpc": 1.0, "function": function, "data": data}



from flask import Blueprint

from flask_rpc.latest import RPC
from .funcs import add_numbers, add_string

bp = Blueprint("tester", __name__, url_prefix="/tester")

rpc = RPC(bp)

rpc.functions(
    add_string=add_string,
)
rpc.functions_auto_name((add_numbers,))



from flask import session

from flask_rpc.latest import RPCResponse


# _ is used to show that the function does not use the request object
def get_session(_):
    return RPCResponse.success(data={**session})



import typing as t


class RPCResponse:
    @classmethod
    def fail(
        cls,
        message: str = None,
        data: t.Union[
            str, int, float, bool, t.List[t.Any], t.Dict[str, t.Any], None
        ] = None,
    ):
        """
        Return a failed response.

        Version 1.0.

        :param message: Str
        :param data: Any (JSON serializable)
        :return:
        """
        r = {"wrpc": 1.0, "ok": False}

        if message:
            r["message"] = message
        if data:
            r["data"] = data

        return r

    @classmethod
    def success(
        cls,
        data: t.Union[
            str, int, float, bool, t.List[t.Any], t.Dict[str, t.Any], None
        ] = None,
        message: str = None,
    ):
        """
        Return a successful response.

        Version 1.0.

        :param data: Any (JSON serializable)
        :param message: Str
        :return:
        """
        r = {"wrpc": 1.0, "ok": True}

        if message:
            r["message"] = message
        if data:
            r["data"] = data

        return r



from flask import Blueprint

from flask_rpc.latest import RPC
from flask_rpc.version_1_0 import RPCAuthSessionKey
from .funcs.login import login
from .funcs.logout import logout
from .funcs.session import get_session

auth = Blueprint("auth", __name__, url_prefix="/auth")

login_rpc = RPC(auth, url_prefix="/login")
login_rpc.functions(login=login)

authenticated_rpc = RPC(auth)
authenticated_rpc.functions(logout=logout, session=get_session)
authenticated_rpc.session_auth(
    RPCAuthSessionKey("logged_in", [True])
)



import typing as t
from datetime import datetime

from sqlalchemy import select

from app.extensions import db


class Clients(db.Model):
    client_id = db.Column(db.Integer, primary_key=True)
    name = db.Column(db.String(50), nullable=False)
    created_at = db.Column(db.DateTime, default=datetime.now())
    updated_at = db.Column(db.DateTime, default=datetime.now(), onupdate=datetime.now())

    @classmethod
    def create(cls, name: str, **_) -> t.Tuple[bool, str, t.Any]:
        client = cls(name=name)

        db.session.add(client)
        db.session.commit()

        return True, "OK", client

    @classmethod
    def read(cls, client_id: int, **_) -> t.Tuple[bool, str, t.Any]:
        q = select(cls).where(cls.client_id == client_id)
        r = db.session.execute(q).scalar_one_or_none()

        if not r:
            return False, "Client not found.", None

        return True, "OK", r

    @classmethod
    def update(cls, client_id: int, name: str, **_) -> t.Tuple[bool, str, t.Any]:
        q = select(cls).where(cls.client_id == client_id)
        r = db.session.execute(q).scalar_one_or_none()

        if not r:
            return False, "Client not found.", None

        r.name = name
        db.session.commit()

        return True, "OK", r

    @classmethod
    def delete(cls, client_id: int, **_) -> t.Tuple[bool, str, t.Any]:
        q = select(cls).where(cls.client_id == client_id)
        r = db.session.execute(q).scalar_one_or_none()

        if not r:
            return False, "Client not found.", None

        db.session.delete(r)
        db.session.commit()

        return True, "OK", r



import typing as t


class RPCAuthSessionKey:
    _key: str
    _value_in: t.List[t.Any]

    def __init__(self, key: str, value_in: t.List[t.Any]):
        self._key = key
        self._value_in = value_in

    def check(self, ask: t.Any):
        if self._key not in ask:
            return False

        if ask.get(self._key) in self._value_in:
            return True

        return False



from flask import Blueprint

from flask_rpc.latest import RPC
from .funcs import create_client, read_client, update_client, delete_client

clients = Blueprint("clients", __name__, url_prefix="/clients")

rpc = RPC(clients)

rpc.functions(
    create=create_client,
    read=read_client,
    update=update_client,
    delete=delete_client,
)



from flask_rpc.latest import RPCResponse


# _ is used to show that the function does not use the request object
def logout(_):
    return RPCResponse.success(
        data={
            "user_id": 1,
        }
    )



from .auth_session_key import RPCAuthSessionKey
from .model import RPCModel
from .request import RPCRequest
from .response import RPCResponse
from .rpc import RPC

__all__ = ["RPC", "RPCResponse", "RPCModel", "RPCRequest", "RPCAuthSessionKey"]



from pprint import pprint

import click
import faker
import requests

from flask_rpc.latest import RPCRequest as Req


@click.group("run")
def run():
    pass


@run.command("create")
def create():
    f = faker.Faker()

    click.echo("Creating a client...")
    response = requests.post(
        "http://127.0.0.1:5000/rpc/clients",
        json={
            "frpc": 1.0,  # Required
            "function": "create",
            "data": {
                "name": f.name()
            }
        }
    )
    pprint(response.json(), indent=2)


# Using the RCPRequest class to make a request to the server

@run.command("create-class")
def create_class():
    f = faker.Faker()

    click.echo("Creating a client...")
    response = requests.post(
        "http://127.0.0.1:5000/rpc/clients",
        json=Req.build(
            function="create",
            data={"name": f.name()}
        )
    )
    pprint(response.json(), indent=2)


@run.command("read")
def read():
    click.echo("Reading client...")
    response = requests.post(
        "http://127.0.0.1:5000/rpc/clients",
        json=Req.build(
            function="read",
            data={"client_id": 1}
        )
    )
    pprint(response.json(), indent=2)


@run.command("read-fail")
def read_fail():
    click.echo("Reading client...")
    response = requests.post(
        "http://127.0.0.1:5000/rpc/clients/read",
        json=Req.build(
            function="read",
            data={"client_id": 11111}
        )
    )
    pprint(response.json(), indent=2)


@run.command("update")
def update():
    click.echo("Updating client...")
    response = requests.post(
        "http://127.0.0.1:5000/rpc/clients/update",
        json=Req.build(
            function="update",
            data={"client_id": 1, "name": "John Doe"}
        )
    )
    pprint(response.json(), indent=2)


@run.command("delete")
def delete():
    click.echo("Deleting client...")
    response = requests.post(
        "http://127.0.0.1:5000/rpc/clients",
        json=Req.build(
            function="delete",
            data={"client_id": 1}
        )
    )
    pprint(response.json(), indent=2)


@run.command("fail")
def fail():
    click.echo("Sending bad command...")
    response = requests.post(
        "http://127.0.0.1:5000/rpc/clients",
        json=Req.build(
            function="fail",
            data={"client_id": 1}
        )
    )
    pprint(response.json(), indent=2)


@run.command("t1")
def test1():
    click.echo("Running function...")
    response = requests.post(
        "http://127.0.0.1:5000/rpc/tester",
        json=Req.build(
            function="add_numbers",
            data=[1, 2, 3]
        )
    )
    pprint(response.json(), indent=2)


@run.command("t2")
def test2():
    click.echo("Running function...")
    response = requests.post(
        "http://127.0.0.1:5000/rpc/tester",
        json=Req.build(
            function="add_string",
            data="World"
        )
    )
    pprint(response.json(), indent=2)


if __name__ == '__main__':
    run()



import typing as t

from flask import Blueprint, Flask, request, session
from pydantic import ValidationError

from ._protocols import RPCAuthSessionKey
from .model import RPCModel
from .response import RPCResponse
from .utilities import snake_case


class RPC:
    LOOKUP: t.Dict[str, t.Callable]

    _host_auth: t.List[str]
    _session_auth: t.Union[RPCAuthSessionKey, t.List[RPCAuthSessionKey]]

    def __init__(
        self,
        app_or_blueprint: t.Union[Flask, Blueprint],
        functions: t.Optional[t.Dict[str, t.Callable]] = None,
        url_prefix: str = "/",
        host_auth: t.Optional[t.List[str]] = None,
        session_auth: t.Optional[
            t.Union[RPCAuthSessionKey, t.List[RPCAuthSessionKey]]
        ] = None,
    ):
        """
        Register the RPC route.

        :param app_or_blueprint: Flask / Blueprint
        :param url_prefix: Str
        """
        self.LOOKUP = {}

        if not hasattr(app_or_blueprint, "add_url_rule"):
            raise TypeError(
                f"Looks like {app_or_blueprint}, type({type(app_or_blueprint)}) might "
                f"not be an instance of Flask, Flask Blueprint or be compatible with "
                "setting Flask routes."
            )

        self._register_route(app_or_blueprint, url_prefix)

        if functions:
            self.functions(**functions)

        if host_auth:
            self.host_auth(host_auth)
        else:
            self.host_auth([])

        if session_auth:
            self.session_auth(session_auth)
        else:
            self.session_auth([])

    def host_auth(self, hosts: t.List[str]):
        self._host_auth = hosts

    def session_auth(
        self, auth_session_keys: t.Union[RPCAuthSessionKey, t.List[RPCAuthSessionKey]]
    ):
        self._session_auth = auth_session_keys

    def functions(self, **kwargs: t.Callable):
        """
        Register RPC functions.

        remote_name=local_name

        :param kwargs:
        :return: None
        """
        for k, v in kwargs.items():
            if not callable(v):
                raise TypeError(f"Expected a callable, got {type(v)}.")

            if not v.__name__:
                raise ValueError(f"Callable {v} must have a name.")

            if k in self.LOOKUP:
                raise ValueError(f"Function {k} already exists.")

            self.LOOKUP[k] = v

    def functions_auto_name(self, functions: t.Iterable[t.Callable]):
        """
        Register RPC functions with their local names.

        remote_name=local_name

        remote_name will always be the name of the function.

        :param functions: Iterable of functions
        :return: None
        """
        for f in functions:
            if not callable(f):
                raise TypeError(f"Expected a callable, got {type(f)}.")

            if not f.__name__:
                raise ValueError(f"Callable {f} must have a name.")

            if f.__name__ in self.LOOKUP:
                raise ValueError(f"Function {f.__name__} already exists.")

            self.LOOKUP[f.__name__] = f

    def _register_route(
        self, route_compatible: t.Union[Flask, Blueprint], url_prefix: str
    ):
        if not url_prefix.startswith("/"):
            url_prefix = f"/{url_prefix}"

        _default = ("/", "/rpc")
        _blueprint_name = ""
        if isinstance(route_compatible, Blueprint):
            _blueprint_name = f"_{route_compatible.name}"

        route_compatible.add_url_rule(
            url_prefix,
            view_func=self._rpc_route,
            endpoint="_rpc"
            if url_prefix in _default
            else f"_rpc{_blueprint_name}_{snake_case(url_prefix)}",
            provide_automatic_options=True,
            methods=["POST"],
        )

    def _rpc_route(self):
        if not self.LOOKUP:
            return RPCResponse.fail("No functions registered.")

        if self._session_auth:
            if isinstance(self._session_auth, RPCAuthSessionKey):
                if not self._session_auth.check(session):
                    return RPCResponse.fail("Unauthorized.")

            if isinstance(self._session_auth, list):
                for auth_session_key in self._session_auth:
                    if isinstance(auth_session_key, RPCAuthSessionKey):
                        if not auth_session_key.check(session):
                            return RPCResponse.fail("Unauthorized.")
                    else:
                        raise ValueError("Invalid session_auth type.")

        if self._host_auth:
            if request.host not in self._host_auth:
                return RPCResponse.fail(f"Unauthorized ({request.host})")

        if not request.is_json:
            return RPCResponse.fail("Request must be JSON.")

        _json = request.get_json()

        if not _json:
            return RPCResponse.fail("Request must not be empty.")

        if not _json.get("wrpc") == 1.0:
            return RPCResponse.fail("Invalid wrpc version.")

        try:
            rpcm = RPCModel(**_json)
        except ValidationError:
            return RPCResponse.fail("Invalid request.")

        try:
            assert rpcm.function in self.LOOKUP
        except AssertionError:
            return RPCResponse.fail("Invalid function.")

        if successful_response := self.LOOKUP[rpcm.function](rpcm.data):
            return successful_response

        return RPCResponse.fail("Unsuccessful command execution.")



from app.models.clients import Clients
from flask_rpc.latest import RPCResponse


def create_client(data):
    status, message, result = Clients.create(**data)

    if not status:
        return RPCResponse.fail(message, data)

    return RPCResponse.success(
        data={
            "client_id": result.client_id,
            "name": result.name,
            "created_at": result.created_at,
            "updated_at": result.updated_at,
        }
    )


def read_client(data):
    status, message, result = Clients.read(**data)

    if not status:
        return RPCResponse.fail(message, data)

    return RPCResponse.success(
        data={
            "client_id": result.client_id,
            "name": result.name,
            "created_at": result.created_at,
            "updated_at": result.updated_at,
        }
    )


def update_client(data):
    status, message, result = Clients.update(**data)

    if not status:
        return RPCResponse.fail(message, data)

    return RPCResponse.success(
        data={
            "client_id": result.client_id,
            "name": result.name,
            "created_at": result.created_at,
            "updated_at": result.updated_at,
        }
    )


def delete_client(data):
    status, message, result = Clients.delete(**data)

    if not status:
        return RPCResponse.fail(message, data)

    return RPCResponse.success(
        {"client_id": result.client_id, "name": result.name}
    )



from flask import session

from flask_rpc.latest import RPCResponse


# _ is used to show that the function does not use the request object
def login(_):
    session["user_id"] = 100
    return RPCResponse.success(
        data={
            "user_id": 100,
        }
    )
